!Hybrid MPI/OpenMP program to calculate PI
! With MIC Offload directives
program pi_hybrid
  use mpi
  implicit none 

  integer*8, parameter ::  nsize=201000000
  integer              ::  nprocs, myrank, ierr
  real*8               ::  mypi,pi

  call mpi_init(ierr)
  call mpi_comm_size(mpi_comm_world,nprocs,ierr)
  call mpi_comm_rank(mpi_comm_world,myrank,ierr)

  write(*,*) 'My rank is: ',myrank,'/',nprocs

  call mpi_barrier(mpi_comm_world,ierr)

  if(myrank == 0) write(*,*)

  !Compute pi = integral(4/(1+x2) on the interval (0,1)
  call do_some_integratin(mypi,nsize,myrank,nprocs)

  !Collect all the partial sums
  call MPI_REDUCE(mypi,pi,1,MPI_DOUBLE_PRECISION,MPI_SUM,0, &
              MPI_COMM_WORLD,ierr)

  call mpi_barrier(mpi_comm_world,ierr)

  if (myrank == 0) then
    write(*,'(a,f11.8)') 'The value of pi is: ',pi
  endif

  call mpi_finalize(ierr)

stop
end

!Integration subroutine that uses OpenMP
subroutine do_some_integratin(mypi,nsize,myrank,nprocs)

  use omp_lib
  implicit none

  integer       ::  myrank, nprocs
  integer*8     ::  nsize, i, start_int, end_int
  integer       ::  iam, np
  real*8        ::  mypi, h, x
  real*8        ::  sum = 0d0

  !Calculate the interval size
  h = 1.0d0/(nsize*nprocs)
 
  !Calculate the interval section
  start_int = nsize*myrank + 1
  end_int = start_int + nsize

  !Begin the OpenMP section
!dir$ offload begin target (mic:0) in(start_int,end_int) out(sum)
!$omp parallel private(iam,x,i,np)
!$omp barrier
   iam = omp_get_thread_num()
   np=omp_get_num_threads()

   !Output the # of tasks and threads
   !Comment this out to reduce output
   write(6,110) iam, myrank, np,nprocs
 110 format('Thread:',i5,' rank:',i5,' threads:',i5,' processes:',i5)  
!$omp barrier

  !Integrate over the appropriate interval over np threads
!$omp do schedule(static) &
!$omp reduction(+:sum)
  do i = start_int, end_int
    x = h * (dble(i) - 0.5d0)
    sum = sum + (4.d0 / (1.d0 + x*x))
 enddo
!$omp end do
!$omp end parallel
!dir$  end offload

  mypi = h * sum

return
end
